/* You may find the license in the LICENSE file */

const EXPORTED_SYMBOLS = [
	'prealloc'
];
	
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;
const module = Cu.import;
const Exception = Components.Exception;

const FileOutputStream = Components.Constructor('@mozilla.org/network/file-output-stream;1', 'nsIFileOutputStream', 'init');
const File = Components.Constructor('@mozilla.org/file/local;1', 'nsILocalFile', 'initWithPath');

// Minimum size of a preallocation.
// If requested size is less then no actual pre-allocation will be performed.
let SIZE_MIN = 5 * 1024 * 1024;

// Step size of the allocation
// Do this step wise to avoid certain "sparse files" cases
const SIZE_STEP = 100 * 1024 * 1024;

module('resource://dta/cothread.jsm');
module('resource://dta/utils.jsm');
module('resource://dta/version.jsm');

// Store workers here.
// Not storing workers (in this context) will cause gc havoc.
const workers = {};

/**
 * Pre-allocates a given file on disk
 * and calls given callback when done
 * 
 * @param file (nsIFile) file to allocate
 * @param size (int) Size to allocate  
 * @param perms (int) *nix file permissions
 * @param callback (function) Callback called once done
 * @param tp (function) Scope (this) to call the callback function in 
 * @return (nsICancelable) Pre-allocation object.
 */
function prealloc(file, size, perms, callback, tp) {
	tp = tp || null;
	if (size <= SIZE_MIN || !isFinite(size)) {
		Debug.log("pa: not preallocating");
		if (callback) {
			callback.call(tp, false);
		}
		return null;
	}
	
	return new WorkerJob(file.path, size, perms, callback, tp);
}

function WorkerJob(path, size, perms, callback, tp) {
	this.path = path;
	this.size = size;
	this.perms = perms;
	this.callback = callback;
	this.tp = tp;
	this.uuid = newUUIDString();
	
	// Create thread and dispatch
	let tm = Cc['@mozilla.org/thread-manager;1'].getService(Ci.nsIThreadManager);
	this.thread = tm.newThread(0);
	try {
		let tp = this.thread.QueryInterface(Ci.nsISupportsPriority);
		tp.priority = Ci.nsISupportsPriority.PRIORITY_LOWEST;
	}
	catch (ex) {
		// no op
	}
	this.main = tm.mainThread;
	workers[this.uuid] = this;
	this.thread.dispatch(this, this.thread.DISPATCH_NORMAL);
}

WorkerJob.prototype = {
	QueryInterface: function worker_QueryInterface(iid) {
		if (iid.equals(Ci.nsISupports) || iid.equals(Ci.nsIRunnable) || iid.equals(Ci.nsICancelable)) {
			return this;
		}
		throw Cr.NS_ERROR_NO_INTERFACE;
	},
	run: function worker_run() {
		let rv = false;
		try {
			let file = new File(this.path);
			let stream = new FileOutputStream(file, 0x02 | 0x08, this.perms, 0);
			try {
				let seekable = stream.QueryInterface(Ci.nsISeekableStream);
				seekable.seek(0x02, 0);
				let i = seekable.tell() + SIZE_STEP;
				for (i = Math.min(this.size - 1, i); !this.terminated && i < this.size - 1; i = Math.min(this.size - 1, i + SIZE_STEP)) {
					seekable.seek(0x00, i);
					stream.write("a", 1);
				}
				rv = true;				
			}
			catch (iex) {
				Debug.log("pa: Failed to run prealloc loop", iex);
			}
			stream.close();
		}
		catch (ex) {
			Debug.log("pa: Failed to run prealloc worker", ex);
		}
		
		// Dispatch event back to the main thread
		this.main.dispatch(new MainJob(this.uuid, this.thread, this.callback, this.tp, rv), this.main.DISPATCH_NORMAL);		
	},
	cancel: function() {
		Debug.log("pa: cancel called!");
		this.terminated = true;
		this.thread.shutdown();
	}
};

if (Version.OS == 'winnt') {
	SIZE_MIN = 30 * 1024;
 	WorkerJob.prototype.run = function workerwin_run() {
		let rv = false;
		try {
			let file = new File(this.path);
			let stream = new FileOutputStream(file, 0x02 | 0x08, this.perms, 0);
			try {
				let seekable = stream.QueryInterface(Ci.nsISeekableStream);
				seekable.seek(0x02, this.size);
				seekable.setEOF();
			}
			finally {
				stream.close();
			}
			rv = true;
		}
		catch (ex) {
			Debug.log("pa: Failed to run prealloc worker", ex);
		}
		
		// Dispatch event back to the main thread
		this.main.dispatch(new MainJob(this.uuid, this.thread, this.callback, this.tp, rv), this.main.DISPATCH_NORMAL);		
	}
}

function MainJob(uuid, thread, callback, tp, result) {
	this.uuid = uuid;
	this.thread = thread;
	this.callback = callback;
	this.tp = tp;
	this.result = result;
}
MainJob.prototype = {
	QueryInterface: WorkerJob.prototype.QueryInterface,
	
	run: function main_run() {
		// thread is done
	
		try {
			// wait for thread to actually join, if not already joined
			this.thread.shutdown();
		}
		catch (ex) {
			// might throw; see Worker.cancel
		}
		
		if (this.callback) {
			try {
				// call the user callback
				this.callback.call(this.tp, this.result);
				Debug.log("pa: prealloc done");
			}
			catch (ex) {
				Debug.log("pa: callback throw", ex);
			}
		}
		
		// cleanup
		workers[this.uuid] = 0;
		delete workers[this.uuid];
	}
};